#!/usr/bin/env python

import logging as log
import os

from fameio.source.cli import (
    Config,
    arg_handling_convert_results,
    get_config_or_default,
)
from fameio.source.logs import log_and_raise_critical, set_up_logger
from fameio.source.reader import ProtoReader

DEFAULT_CONFIG = {
    Config.LOG_LEVEL: "info",
    Config.LOG_FILE: None,
    Config.AGENT_LIST: None,
    Config.OUTPUT: None,
    Config.SINGLE_AGENT_EXPORT: False,
}


def build_output_folder_name(config_output: str, input_file_path: str) -> str:
    """Returns the name of the output folder - derived either from the specified `config_output` or `input_file_path`"""
    if config_output:
        log.info("Using specified output path: {}".format(config_output))
        output_folder_name = config_output
    else:
        file_name_without_folder = input_file_path.split("\\")[-1]
        output_folder_name = file_name_without_folder.replace(".pb", "")
        log.info("No output path specified - writing to: {}".format(output_folder_name))
    return output_folder_name


def write_csv_files(output_folder: str, agent_data: dict, single_export: bool) -> None:
    """Writes given `agent_data` to .csv files at given `output_folder` grouping for `agent_id` if enabled"""
    for class_name, class_data in agent_data.items():
        class_data.sort_index(inplace=True)
        if single_export:
            for agent_id, data in class_data.groupby("AgentId"):
                out_file_name = (
                    output_folder + "/" + class_name + "_" + str(agent_id) + ".csv"
                )
                data.to_csv(out_file_name, sep=";", header=True, index=True)
        else:
            out_file_name = output_folder + "/" + class_name + ".csv"
            class_data.to_csv(out_file_name, sep=";", header=True, index=True)


def run(file_path: str, config: dict = None) -> None:
    """Reads file in protobuf format at given `file_path` and extracts its content to .csv file(s)"""
    config = get_config_or_default(config, DEFAULT_CONFIG)
    set_up_logger(
        level_name=config[Config.LOG_LEVEL], file_name=config[Config.LOG_FILE]
    )

    log.info("Reading protobuffer file...")
    proto_reader = ProtoReader(file_path)
    if not proto_reader.has_agent_data():
        log_and_raise_critical(
            "Provided protobuf file contains no data in output section."
        )

    log.info("Extracting agent data...")
    agent_data = proto_reader.extract_agent_data(config[Config.AGENT_LIST])

    output_folder = build_output_folder_name(config[Config.OUTPUT], file_path)
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    log.info("Writing data to .csv files...")
    write_csv_files(output_folder, agent_data, config[Config.SINGLE_AGENT_EXPORT])
    log.info("Data conversion completed.")


if __name__ == "__main__":
    input_file, run_config = arg_handling_convert_results(DEFAULT_CONFIG)
    run(input_file, run_config)



import logging as log
from collections import Counter
from typing import Any, Dict, List

from fameio.source.logs import log_error_and_raise
from fameio.source.scenario import Agent, Attribute, Contract, Scenario
from fameio.source.schema.agenttype import AgentType
from fameio.source.schema.attribute import AttributeSpecs, AttributeType
from fameio.source.schema.schema import Schema
from fameio.source.time import FameTime


class ValidationException(Exception):
    """Indicates an error occurred during validation of any data with a connected schema"""

    pass


class SchemaValidator:
    """Handles validation of scenarios based on a connected `schema`"""

    _AGENT_ID_NOT_UNIQUE = "Agent ID(s) not unique: '{}'."
    _AGENT_TYPE_UNKNOWN = "Agent type '{}' not declared in Schema."
    _ATTRIBUTE_UNKNOWN = "Attribute '{}' not declared in Schema."
    _TYPE_NOT_IMPLEMENTED = "Validation not implemented for AttributeType '{}'."
    _INCOMPATIBLE = "Value '{}' incompatible with {} of Attribute '{}'."
    _DISALLOWED = "Value '{}' not in list of allowed values of Attribute '{}'"
    _AGENT_MISSING = "Contract mentions Agent with ID '{}' but Agent was not declared."
    _PRODUCT_MISSING = "Product '{}' not declared in Schema for AgentType '{}'."
    _KEY_MISSING = "Required key '{}' missing in dictionary '{}'."
    _ATTRIBUTE_MISSING = "Mandatory attribute '{}' missing."
    _DEFAULT_IGNORED = "Optional Attribute: '{}': not specified - provided Default ignored for optional Attributes."
    _OPTIONAL_MISSING = "Optional Attribute: '{}': not specified."
    _IS_NO_LIST = "Attribute '{}' is list but assigned value '{}' is not a list."

    @staticmethod
    def _ensure_mandatory_present(
        attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs]
    ) -> None:
        """
        Raises Exception if in given list of `specifications` at least one item is mandatory,
        provides no defaults and is not contained in given `attributes` dictionary
        """
        for name, specification in specifications.items():
            if name not in attributes:
                if specification.is_mandatory:
                    if not specification.has_default():
                        log_error_and_raise(
                            ValidationException(
                                SchemaValidator._ATTRIBUTE_MISSING.format(name)
                            )
                        )
                else:
                    if specification.has_default():
                        log.warning(SchemaValidator._DEFAULT_IGNORED.format(name))
                    else:
                        log.info(SchemaValidator._OPTIONAL_MISSING.format(name))
            if name in attributes and specification.has_nested_attributes():
                attribute = attributes[name]
                if specification.is_list:
                    for entry in attribute.nested_list:
                        SchemaValidator._ensure_mandatory_present(
                            entry, specification.get_attributes()
                        )
                else:
                    SchemaValidator._ensure_mandatory_present(
                        attribute.nested, specification.get_attributes()
                    )

    @staticmethod
    def _get_agent(schema: Schema, name: str) -> AgentType:
        """Returns agent specified by `name` or raises Exception if this agent is not present in given `schema`"""
        if name in schema.agent_types:
            return schema.agent_types[name]
        else:
            log_error_and_raise(
                ValidationException(SchemaValidator._AGENT_TYPE_UNKNOWN.format(name))
            )

    @staticmethod
    def ensure_is_valid_agent(agent: Agent, schema: Schema) -> None:
        """Raises an exception if given `agent` does not meet the specified `schema` requirements"""
        scenario_attributes = agent.attributes
        schema_attributes = SchemaValidator._get_agent(
            schema, agent.type_name
        ).attributes
        SchemaValidator._ensure_mandatory_present(
            scenario_attributes, schema_attributes
        )
        SchemaValidator._ensure_attributes_exist(scenario_attributes, schema_attributes)
        SchemaValidator._ensure_value_matches_type(
            scenario_attributes, schema_attributes
        )

    @staticmethod
    def _ensure_attributes_exist(
        attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs]
    ) -> None:
        """Raises exception any entry of given `attributes` has no corresponding type `specification`"""
        for name, attribute in attributes.items():
            if name not in specifications:
                log_error_and_raise(
                    ValidationException(
                        SchemaValidator._ATTRIBUTE_UNKNOWN.format(attribute)
                    )
                )
            if attribute.has_nested:
                specification = specifications[name]
                SchemaValidator._ensure_attributes_exist(
                    attribute.nested, specification.get_attributes()
                )
            if attribute.has_nested_list:
                specification = specifications[name]
                for entry in attribute.nested_list:
                    SchemaValidator._ensure_attributes_exist(
                        entry, specification.get_attributes()
                    )

    @staticmethod
    def _ensure_value_matches_type(
        attributes: Dict[str, Attribute], specifications: Dict[str, AttributeSpecs]
    ) -> None:
        """Raises exception if in given list of `attributes` its value does not match associated type `specification`"""
        for name, attribute in attributes.items():
            specification = specifications[name]
            if attribute.has_value:
                value = attribute.value
                type_spec = specification.get_type()
                if not SchemaValidator._is_compatible(specification, value):
                    message = SchemaValidator._INCOMPATIBLE.format(
                        value, type_spec, specification
                    )
                    log_error_and_raise(ValidationException(message))
                if not SchemaValidator._is_allowed_value(specification, value):
                    log_error_and_raise(
                        ValidationException(
                            SchemaValidator._DISALLOWED.format(value, specification)
                        )
                    )
            if attribute.has_nested:
                SchemaValidator._ensure_value_matches_type(
                    attribute.nested, specification.get_attributes()
                )
            if attribute.has_nested_list:
                for entry in attribute.nested_list:
                    SchemaValidator._ensure_value_matches_type(
                        entry, specification.get_attributes()
                    )

    @staticmethod
    def _is_compatible(specification: AttributeSpecs, value_or_values: Any) -> bool:
        """Returns True if given `value_or_values` is compatible to specified `attribute_type` and `should_be_list`"""
        is_list = isinstance(value_or_values, list)
        attribute_type = specification.get_type()
        if specification.is_list:
            if not is_list:
                log.warning(
                    SchemaValidator._IS_NO_LIST.format(
                        specification.get_full_name(), value_or_values
                    )
                )
                return SchemaValidator._is_compatible_value(
                    attribute_type, value_or_values
                )
            for value in value_or_values:
                if not SchemaValidator._is_compatible_value(attribute_type, value):
                    return False
            return True
        else:
            return (not is_list) and SchemaValidator._is_compatible_value(
                attribute_type, value_or_values
            )

    @staticmethod
    def _is_compatible_value(attribute_type: AttributeType, value) -> bool:
        """Returns True if given single value is compatible to specified `attribute_type`"""
        if attribute_type is AttributeType.INTEGER:
            if isinstance(value, int):
                return -2147483648 < value < 2147483647
            return False
        if attribute_type is AttributeType.LONG:
            return isinstance(value, int)
        elif attribute_type is AttributeType.DOUBLE:
            return isinstance(value, (int, float))
        elif attribute_type in (AttributeType.ENUM, AttributeType.STRING):
            return isinstance(value, str)
        elif attribute_type is AttributeType.TIME_STAMP:
            return FameTime.is_fame_time_compatible(value)
        elif attribute_type is AttributeType.TIME_SERIES:
            return isinstance(value, (str, int, float))
        else:
            log_error_and_raise(
                ValidationException(
                    SchemaValidator._TYPE_NOT_IMPLEMENTED.format(attribute_type)
                )
            )

    @staticmethod
    def _is_allowed_value(attribute: AttributeSpecs, value) -> bool:
        """Returns True if `value` matches an entry of given `Attribute`'s value list or if this list is empty"""
        if not attribute.values:
            return True
        else:
            return value in attribute.values

    @staticmethod
    def ensure_is_valid_contract(
        contract: Contract, schema: Schema, agent_types_by_id: Dict[int, str]
    ) -> None:
        """Raises exception if given `contract` does not meet the `schema`'s requirements, using `agent_types_by_id`"""
        sender_id = contract.sender_id
        if sender_id not in agent_types_by_id:
            log_error_and_raise(
                ValidationException(SchemaValidator._AGENT_MISSING.format(sender_id))
            )
        if contract.receiver_id not in agent_types_by_id:
            log_error_and_raise(
                ValidationException(
                    SchemaValidator._AGENT_MISSING.format(contract.receiver_id)
                )
            )
        sender_type_name = agent_types_by_id[sender_id]
        if sender_type_name not in schema.agent_types:
            log_error_and_raise(
                ValidationException(
                    SchemaValidator._AGENT_TYPE_UNKNOWN.format(sender_type_name)
                )
            )
        sender_type = schema.agent_types[sender_type_name]
        product = contract.product_name
        if product not in sender_type.products:
            log_error_and_raise(
                ValidationException(
                    SchemaValidator._PRODUCT_MISSING.format(product, sender_type_name)
                )
            )

    @staticmethod
    def ensure_unique_agent_ids(agents: List[Agent]) -> None:
        """Raises exception if any id for given `agents` is not unique"""
        list_of_ids = [agent.id for agent in agents]
        non_unique_ids = [
            agent_id for agent_id, count in Counter(list_of_ids).items() if count > 1
        ]
        if non_unique_ids:
            log_error_and_raise(
                ValidationException(
                    SchemaValidator._AGENT_ID_NOT_UNIQUE.format(non_unique_ids)
                )
            )

    @staticmethod
    def ensure_agent_type_in_schema(agent: Agent, schema: Schema) -> None:
        """Raises if exception if type for given `agent` is not specified in given `schema`"""
        if agent.type_name not in schema.agent_types:
            log_error_and_raise(
                ValidationException(
                    SchemaValidator._AGENT_TYPE_UNKNOWN.format(agent.type_name)
                )
            )

    @staticmethod
    def ensure_is_valid_scenario(scenario: Scenario) -> None:
        """Raises exception if given `scenario` does not meet its own schema requirements"""
        schema = scenario.schema
        agents = scenario.agents

        SchemaValidator.ensure_unique_agent_ids(agents)
        for agent in agents:
            SchemaValidator.ensure_agent_type_in_schema(agent, schema)
            SchemaValidator.ensure_is_valid_agent(agent, schema)

        agent_types_by_id = {agent.id: agent.type_name for agent in agents}
        for contract in scenario.contracts:
            SchemaValidator.ensure_is_valid_contract(
                contract, schema, agent_types_by_id
            )



from typing import Any, Dict

from fameio.source.scenario.attribute import Attribute
from fameio.source.scenario.exception import (
    assert_or_raise,
    get_or_default,
    get_or_raise,
)
from fameio.source.tools import keys_to_lower


class Agent:
    """Contains specifications for an agent in a scenario"""

    _KEY_TYPE = "Type".lower()
    _KEY_ID = "Id".lower()
    _KEY_ATTRIBUTES = "Attributes".lower()

    _MISSING_KEY = "Agent requires `key` '{}' but is missing it."
    _MISSING_TYPE = "Agent requires `type` but is missing it."
    _MISSING_ID = "Agent requires a positive integer `id` but was '{}'."
    _DOUBLE_ATTRIBUTE = (
        "Cannot add attribute '{}' to agent {} because it already exists."
    )

    def __init__(self, agent_id: int, type_name: str) -> None:
        """Constructs a new Agent"""
        assert_or_raise(
            type(agent_id) is int and agent_id >= 0, self._MISSING_ID.format(agent_id)
        )
        assert_or_raise(type_name and len(type_name.strip()) > 0, self._MISSING_TYPE)
        self._id = agent_id
        self._type_name = type_name.strip()
        self._attributes = dict()

    @classmethod
    def from_dict(cls, definitions: dict) -> "Agent":
        """Parses an agent from provided `definitions`"""
        definitions = keys_to_lower(definitions)
        agent_type = get_or_raise(definitions, Agent._KEY_TYPE, Agent._MISSING_KEY)
        agent_id = get_or_raise(definitions, Agent._KEY_ID, Agent._MISSING_KEY)
        result = cls(agent_id, agent_type)
        attribute_definitions = get_or_default(
            definitions, Agent._KEY_ATTRIBUTES, dict()
        )
        result.__init_attributes_from_dict(attribute_definitions)
        return result

    def to_dict(self) -> dict:
        """Serializes the Agent content to a dict"""
        result = dict()

        result[Agent._KEY_TYPE] = self.type_name
        result[Agent._KEY_ID] = self.id

        if len(self.attributes) > 0:
            attributes_dict = dict()
            for attr_name, attr_value in self.attributes.items():
                attributes_dict[attr_name] = attr_value.generic_content
            result[self._KEY_ATTRIBUTES] = attributes_dict

        return result

    def _notify_data_changed(self):
        """Placeholder method used to signal data changes to derived types"""
        pass

    @property
    def id(self) -> int:
        """Returns the ID of the Agent"""
        return self._id

    @property
    def display_id(self) -> str:
        """Returns the ID of the Agent as a string for display purposes"""
        return "#{}".format(self._id)

    @property
    def type_name(self) -> str:
        """Returns the name of the Agent type"""
        return self._type_name

    @property
    def attributes(self) -> Dict[str, Attribute]:
        """Returns dictionary of all Attributes of this agent"""
        return self._attributes

    def add_attribute(self, name: str, value: Attribute):
        """Adds a new attribute to the Agent (raise an error if it already exists)"""
        if name in self._attributes:
            raise ValueError(self._DOUBLE_ATTRIBUTE.format(name, self.display_id))
        self._attributes[name] = value
        self._notify_data_changed()

    def __init_attributes_from_dict(self, attributes: Dict[str, Any]) -> None:
        """Initialize Agent `attributes` from dict; Must only be called when creating a new Agent"""
        assert len(self._attributes) == 0
        self._attributes = {}
        for name, value in attributes.items():
            full_name = str(self.type_name) + "(" + str(self.id) + "): " + name
            self.add_attribute(name, Attribute(full_name, value))



import logging as log
import pathlib as pt
from typing import Any, Dict

import pandas as pd
from fameio.source.logs import log_error_and_raise
from fameio.source.scenario import (
    Agent,
    Attribute,
    Contract,
    GeneralProperties,
    Scenario,
)
from fameio.source.path_resolver import PathResolver
from fameio.source.schema.attribute import AttributeSpecs, AttributeType
from fameio.source.series import TimeSeriesManager
from fameio.source.time import START_IN_REAL_TIME, ConversionException, FameTime
from fameio.source.tools import ensure_is_list
from fameprotobuf import DataStorage_pb2


class ProtoWriterException(Exception):
    """Indicates an error during writing of protobuffer file"""

    pass


class ProtoWriter:
    """Writes a given scenario to protobuffer file"""

    _TYPE_NOT_IMPLEMENTED = "AttributeType '{}' not implemented."
    _NOT_TIME_SERIES = "Cannot convert value '{}' to TimeSeries."
    _CONTRACT_UNSUPPORTED = (
        "Unsupported value for Contract Attribute '{}'; "
        "Only support `int`, `float`, `enum` or `dict` types are supported here."
    )
    _USING_DEFAULT = "Using provided Default for Attribute: '{}'."
    _CORRUPT_TIME_SERIES_VALUE = "TimeSeries file '{}' is corrupt: At least one entry in value column is not numeric."
    _CORRUPT_TIME_SERIES_KEY = "TimeSeries file '{}' is corrupt: At least one entry in first column is not a timestamp."

    def __init__(self, file_path: str, path_resolver=PathResolver()) -> None:
        self.file_path = file_path
        self.time_series_manager = TimeSeriesManager()
        self._path_resolver = path_resolver

    def write_validated_scenario(self, scenario: Scenario) -> None:
        """Writes given validated Scenario to file"""
        log.info("Writing scenario to protobuf file `{}`".format(self.file_path))
        pb_data_storage = DataStorage_pb2.DataStorage()
        pb_input = pb_data_storage.input

        log.info("Adding General Properties")
        ProtoWriter._set_general_properties(pb_input, scenario.general_properties)

        log.info("Adding Agents")
        schema = scenario.schema
        for agent in scenario.agents:
            pb_agent = ProtoWriter._set_agent(pb_input.agent.add(), agent)
            attribute_specs = schema.agent_types[agent.type_name].attributes
            self._set_attributes(pb_agent, agent.attributes, attribute_specs)

        log.info("Adding Contracts")
        for contract in scenario.contracts:
            pb_contract = ProtoWriter._set_contract(pb_input.contract.add(), contract)
            ProtoWriter._set_contract_attributes(pb_contract, contract.attributes)

        log.info("Adding TimeSeries")
        self._set_time_series(pb_input)
        log.info("Writing to disk")
        self._write_protobuf_to_disk(pb_data_storage)

    @staticmethod
    def _set_general_properties(pb_input, gen_props: GeneralProperties) -> None:
        """Saves Scenario's general properties to specified protobuffer `pb_input` container"""
        pb_input.runId = gen_props.run_id
        pb_input.simulation.startTime = gen_props.simulation_start_time
        pb_input.simulation.stopTime = gen_props.simulation_stop_time
        pb_input.simulation.randomSeed = gen_props.simulation_random_seed
        pb_input.output.interval = gen_props.output_interval
        pb_input.output.process = gen_props.output_process

    @staticmethod
    def _set_agent(pb_agent, agent: Agent):
        """Saves type and id of given `agent` to protobuffer `pb_agent` container. Returns given `pb_agent`."""
        pb_agent.className = agent.type_name
        pb_agent.id = agent.id
        return pb_agent

    def _set_attributes(
        self,
        pb_parent,
        attributes: Dict[str, Attribute],
        specs: Dict[str, AttributeSpecs],
    ) -> None:
        """Assigns `attributes` to protobuffer fields of given `pb_parent` - cascades for nested Attributes"""
        values_not_set = [key for key in specs.keys()]
        for name, attribute in attributes.items():
            pb_field = ProtoWriter._add_field(pb_parent, name)
            attribute_specs = specs[name]
            values_not_set.remove(name)
            attribute_type = attribute_specs.get_type()
            if attribute_type is AttributeType.BLOCK:
                if attribute_specs.is_list:
                    for index, entry in enumerate(attribute.nested_list):
                        pb_inner = ProtoWriter._add_field(pb_field, str(index))
                        self._set_attributes(
                            pb_inner, entry, attribute_specs.get_attributes()
                        )
                else:
                    self._set_attributes(
                        pb_field,
                        attribute.nested,
                        attribute_specs.get_attributes(),
                    )
            else:
                self._set_attribute(pb_field, attribute.value, attribute_type)
        for name in values_not_set:
            attribute_specs = specs[name]
            if attribute_specs.is_mandatory:
                pb_field = ProtoWriter._add_field(pb_parent, name)
                self._set_attribute(
                    pb_field, attribute_specs.default, attribute_specs.get_type()
                )
                log.info(ProtoWriter._USING_DEFAULT.format(name))

    @staticmethod
    def _add_field(pb_parent, name: str) -> Any:
        """Returns new field with given `name` that is added to given `pb_parent`"""
        pb_field = pb_parent.field.add()
        pb_field.fieldName = name
        return pb_field

    def _set_attribute(self, pb_field, value, attribute_type: AttributeType) -> None:
        """Sets given `value` to given protobuffer `pb_field` depending on specified `attribute_type`"""
        if attribute_type is AttributeType.INTEGER:
            pb_field.intValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.DOUBLE:
            pb_field.doubleValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.LONG:
            pb_field.longValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.TIME_STAMP:
            pb_field.longValue.extend(
                ensure_is_list(FameTime.convert_string_if_is_datetime(value))
            )
        elif attribute_type in (AttributeType.ENUM, AttributeType.STRING):
            pb_field.stringValue.extend(ensure_is_list(value))
        elif attribute_type is AttributeType.TIME_SERIES:
            self._set_time_series_from_value(pb_field, value)
        else:
            log_error_and_raise(
                ProtoWriterException(
                    ProtoWriter._TYPE_NOT_IMPLEMENTED.format(attribute_type)
                )
            )

    def _set_time_series_from_value(self, pb_field, value):
        """Hands given `value` to TimeSeriesManager to assign a unique id which is then set to `pb_field`.seriesId"""
        if not isinstance(value, (str, int, float)):
            log_error_and_raise(
                ProtoWriterException(ProtoWriter._NOT_TIME_SERIES.format(str(value)))
            )
        elif isinstance(value, str):
            value = pt.Path(value).as_posix()
        pb_field.seriesId = self.time_series_manager.save_get_time_series_id(value)

    def _set_time_series(self, pb_input):
        """Adds all time series from TimeSeriesManager to given `pb_input`"""
        ids_of_series_by_name = self.time_series_manager.get_ids_of_series_by_name()
        for identifier, unique_id in ids_of_series_by_name.items():
            pb_series = pb_input.timeSeries.add()
            pb_series.seriesId = unique_id
            series_name, data_frame = self._get_series_as_dataframe(identifier)
            pb_series.seriesName = series_name
            try:
                ProtoWriter._add_rows_to_series(pb_series, data_frame)
            except TypeError:
                log_error_and_raise(
                    ProtoWriterException(
                        ProtoWriter._CORRUPT_TIME_SERIES_VALUE.format(identifier)
                    )
                )
            except ConversionException:
                log_error_and_raise(
                    ProtoWriterException(
                        ProtoWriter._CORRUPT_TIME_SERIES_KEY.format(identifier)
                    )
                )

    def _get_series_as_dataframe(self, identifier) -> (str, pd.DataFrame):
        """Returns a DataFrame containing the series obtained from the given `identifier` and an associated name"""
        if isinstance(identifier, str):
            # expect the string to be a file path
            series_path = self._path_resolver.resolve_series_file_path(identifier)
            return identifier, pd.read_csv(series_path, sep=";", header=None)
        else:
            name = "Constant value: {}".format(identifier)
            return name, pd.DataFrame(
                {"time": [START_IN_REAL_TIME], "value": [identifier]}
            )

    @staticmethod
    def _add_rows_to_series(series, data_frame):
        for (key, value) in data_frame.itertuples(index=False):
            row = series.row.add()
            row.timeStep = FameTime.convert_string_if_is_datetime(key)
            row.value = value

    def _write_protobuf_to_disk(self, pb_data_storage) -> None:
        """Writes given `protobuf_input_data` to disk"""
        with open(self.file_path, "wb") as file:
            file.write(pb_data_storage.SerializeToString())
        log.info("Saved protobuf file `{}` to disk".format(self.file_path))

    @staticmethod
    def _set_contract(pb_contract, contract: Contract):
        """Saves given `contract` details to protobuffer container `pb_contract`. Returns given `pb_contract`"""
        pb_contract.senderId = contract.sender_id
        pb_contract.receiverId = contract.receiver_id
        pb_contract.productName = contract.product_name
        pb_contract.firstDeliveryTime = contract.first_delivery_time
        pb_contract.deliveryIntervalInSteps = contract.delivery_interval
        if contract.expiration_time:
            pb_contract.expirationTime = contract.expiration_time
        return pb_contract

    @staticmethod
    def _set_contract_attributes(pb_parent, attributes: Dict[str, Attribute]) -> None:
        """Assign (nested) Attributes to given protobuffer container `pb_parent`"""
        for name, attribute in attributes.items():
            log.debug("Assigning contract attribute `{}`.".format(name))
            pb_field = ProtoWriter._add_field(pb_parent, name)

            if attribute.has_value:
                value = attribute.value
                if isinstance(value, int):
                    pb_field.intValue.extend([value])
                elif isinstance(value, float):
                    pb_field.doubleValue.extend([value])
                elif isinstance(value, str):
                    pb_field.stringValue.extend([value])
                else:
                    log_error_and_raise(
                        ProtoWriterException(
                            ProtoWriter._CONTRACT_UNSUPPORTED.format(str(attribute))
                        )
                    )
            elif attribute.has_nested():
                ProtoWriter._set_contract_attributes(pb_field, attribute.nested)



import logging as log

from fameio.source.scenario.exception import get_or_default, get_or_raise
from fameio.source.time import FameTime
from fameio.source.tools import keys_to_lower


class GeneralProperties:
    """Hosts general properties of a scenario"""

    _KEY_RUN = "RunId".lower()
    _KEY_SIMULATION = "Simulation".lower()
    _KEY_START = "StartTime".lower()
    _KEY_STOP = "StopTime".lower()
    _KEY_SEED = "RandomSeed".lower()
    _KEY_OUTPUT = "Output".lower()
    _KEY_INTERVAL = "Interval".lower()
    _KEY_PROCESS = "Process".lower()

    _MISSING_KEY = "General Properties requires key '{}' but it is missing."
    _SIMULATION_DURATION = (
        "Simulation starts after its end time - check start and stop times."
    )

    def __init__(
        self,
        run_id: int,
        simulation_start_time: int,
        simulation_stop_time: int,
        simulation_random_seed: int,
        output_interval: int,
        output_process: int,
    ) -> None:
        if simulation_stop_time < simulation_start_time:
            log.warning(GeneralProperties._SIMULATION_DURATION)
        self._run_id = run_id
        self._simulation_start_time = simulation_start_time
        self._simulation_stop_time = simulation_stop_time
        self._simulation_random_seed = simulation_random_seed
        self._output_interval = output_interval
        self._output_process = output_process

    @classmethod
    def from_dict(cls, definitions: dict) -> "GeneralProperties":
        """Parse general properties from provided `definitions`"""
        definitions = keys_to_lower(definitions)
        run_id = get_or_default(definitions, GeneralProperties._KEY_RUN, 1)

        simulation_definition = keys_to_lower(
            get_or_raise(
                definitions,
                GeneralProperties._KEY_SIMULATION,
                GeneralProperties._MISSING_KEY,
            )
        )
        start_time = FameTime.convert_string_if_is_datetime(
            get_or_raise(
                simulation_definition,
                GeneralProperties._KEY_START,
                GeneralProperties._MISSING_KEY,
            )
        )
        stop_time = FameTime.convert_string_if_is_datetime(
            get_or_raise(
                simulation_definition,
                GeneralProperties._KEY_STOP,
                GeneralProperties._MISSING_KEY,
            )
        )
        random_seed = get_or_default(
            simulation_definition, GeneralProperties._KEY_SEED, 1
        )

        output_definitions = keys_to_lower(
            get_or_default(definitions, GeneralProperties._KEY_OUTPUT, dict())
        )
        output_interval = get_or_default(
            output_definitions, GeneralProperties._KEY_INTERVAL, 100
        )
        output_process = get_or_default(
            output_definitions, GeneralProperties._KEY_PROCESS, 0
        )

        return GeneralProperties(
            run_id, start_time, stop_time, random_seed, output_interval, output_process
        )

    def to_dict(self) -> dict:
        """Serializes the general properties to a dict"""
        result = dict()

        result[self._KEY_RUN] = self.run_id

        simulation_dict = dict()
        simulation_dict[self._KEY_START] = self.simulation_start_time
        simulation_dict[self._KEY_STOP] = self.simulation_stop_time
        simulation_dict[self._KEY_SEED] = self.simulation_random_seed
        result[self._KEY_SIMULATION] = simulation_dict

        output_dict = dict()
        output_dict[self._KEY_INTERVAL] = self.output_interval
        output_dict[self._KEY_PROCESS] = self.output_process
        result[self._KEY_OUTPUT] = output_dict

        return result

    @property
    def run_id(self) -> int:
        """Returns the run ID"""
        return self._run_id

    @property
    def simulation_start_time(self) -> int:
        """Returns the simulation start time"""
        return self._simulation_start_time

    @property
    def simulation_stop_time(self) -> int:
        """Returns the simulation stop time"""
        return self._simulation_stop_time

    @property
    def simulation_random_seed(self) -> int:
        """Returns the simulation random seed"""
        return self._simulation_random_seed

    @property
    def output_interval(self) -> int:
        """Returns the output interval"""
        return self._output_interval

    @property
    def output_process(self) -> int:
        """Returns the output process"""
        return self._output_process



import logging as log
from pathlib import Path
from typing import List

from fameprotobuf.DataStorage_pb2 import DataStorage
from fameprotobuf.Services_pb2 import Output
from pandas import DataFrame


class ProtoReader:
    """Reads protobuffer at given `file_path` and extracts its contents"""

    def __init__(self, file_path: str) -> None:
        self.pb_output = self._read_output_from_proto_file(file_path)

    @staticmethod
    def _read_output_from_proto_file(proto_file_name: str) -> Output:
        """Reads protobuf `DataStorage` of given `proto_file_name` and returns its `output` content"""
        proto_data_storage = DataStorage()
        with open(Path(proto_file_name).as_posix(), "rb") as file:
            proto_data_storage.ParseFromString(file.read())
        return proto_data_storage.output

    def has_agent_data(self):
        """Returns True if data for any agent is present"""
        return len(self.pb_output.agentType) > 0

    def extract_agent_data(self, requested_agents: List[str]) -> dict:
        """Returns dict of DataFrames containing all data of agents listed in `requested_agents` by their class name"""
        agents_to_extract = self._get_agents_to_extract(requested_agents)
        data_frames = dict()
        for agent_type in self.pb_output.agentType:
            class_name = agent_type.className
            if class_name not in agents_to_extract:
                log.info("Ignoring not selected AgentType `{}`".format(class_name))
                continue
            agent_data = self._extract_agent_data(class_name)
            data_frame = DataFrame.from_dict(agent_data, orient="index")
            data_frame.rename(columns=self._get_column_map(agent_type), inplace=True)
            if not data_frame.empty:
                data_frame.rename_axis(("AgentId", "TimeStep"), inplace=True)
                data_frames[class_name] = data_frame
        return data_frames

    def _get_agents_to_extract(self, requested_agents: List[str]) -> List[str]:
        """Returns existing agent types that match `requested_agents` or all available types if given list is emtpy"""
        agents_to_extract = list()
        agent_types = self.pb_output.agentType
        available_agent_types = {
            agent_type.className.upper(): agent_type.className
            for agent_type in agent_types
        }
        if requested_agents:
            for agent in requested_agents:
                if agent.upper() in available_agent_types.keys():
                    agents_to_extract.append(available_agent_types[agent.upper()])
                else:
                    log.error(
                        "No output present for AgentType `{}` - Please check agent type spelling".format(
                            agent
                        )
                    )
        else:
            agents_to_extract = available_agent_types.values()
        return agents_to_extract

    @staticmethod
    def _get_column_map(agent_type) -> dict:
        """Returns dictionary of column IDs mapping to their name"""
        return {field.fieldId: field.fieldName for field in agent_type.field}

    def _extract_agent_data(self, class_name: str) -> dict:
        """Returns dict containing output of all output columns of given `class_name` by agentId and timeStep"""
        series_data = dict()
        for series in self.pb_output.series:
            if class_name == series.className:
                self._add_series_data_to(series, series_data)
        return series_data

    @staticmethod
    def _add_series_data_to(series, container: dict) -> None:
        """Adds data from given protobuffer `series` to specified `container` dict"""
        agent_id = series.agentId
        for line in series.line:
            index = (agent_id, line.timeStep)
            values = {column.fieldId: column.value for column in line.column}
            container[index] = values



import logging as log
from typing import Dict

from fameio.source.schema.attribute import AttributeSpecs
from fameio.source.tools import keys_to_lower


class AgentType:
    """Schema definitions for an Agent type"""

    _NO_ATTRIBUTES = "Agent '{}' has no specified 'Attributes'."
    _NO_PRODUCTS = "Agent '{}' has no specified Products."

    _KEY_ATTRIBUTES = "Attributes".lower()
    _KEY_PRODUCTS = "Products".lower()

    def __init__(self, name: str):
        self._name = name
        self._attributes = dict()
        self._products = list()

    @classmethod
    def from_dict(cls, name: str, definitions: dict) -> "AgentType":
        """Loads an agent type `definition` from the given input dict"""
        definition = keys_to_lower(definitions)

        result = cls(name)
        if AgentType._KEY_ATTRIBUTES in definition:
            for attribute_name, attribute_details in definition[
                AgentType._KEY_ATTRIBUTES
            ].items():
                full_name = name + "." + attribute_name
                result._attributes[attribute_name] = AttributeSpecs(
                    full_name, attribute_details
                )
        else:
            log.info(AgentType._NO_ATTRIBUTES.format(name))

        if AgentType._KEY_PRODUCTS in definition:
            products_to_add = definition[AgentType._KEY_PRODUCTS]
            if isinstance(products_to_add, list):
                result._products.extend(products_to_add)
            else:
                result._products.append(products_to_add)
        else:
            log.info(AgentType._NO_PRODUCTS.format(name))

        return result

    @property
    def name(self) -> str:
        """Returns the agent type name"""
        return self._name

    @property
    def products(self) -> list:
        """Returns list of products or an empty list if no products are defined"""
        return self._products

    @property
    def attributes(self) -> Dict[str, AttributeSpecs]:
        """Returns list of Attributes of this agent or an empty list if no attributes are defined"""
        return self._attributes
